/* play-state-misc.c
 *
 * Copyright 2021 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include "play-state-misc.h"

ClueId
clue_get_id (IPuzCrossword *xword, const IPuzClue *clue)
{
  ClueId id ;

  if (clue)
    {
      id.direction = ipuz_clue_get_direction ((IPuzClue *) clue);
      id.index = ipuz_crossword_find_clue (xword, (IPuzClue *) clue);
    }
  else
    {
      id.direction = IPUZ_CLUE_DIRECTION_NONE;
      id.index = 0;
    }

  return id;
}

IPuzClue *
get_nth_clue_by_direction (IPuzCrossword *xword, IPuzClueDirection direction, guint n)
{
  guint n_clues = ipuz_crossword_get_n_clues (xword, direction);

  if (n < n_clues)
    {
      IPuzClue *clue = ipuz_crossword_get_clue (xword, direction, n);
      g_assert (clue != NULL);
      return clue;
    }
  else
    {
      return NULL;
    }
}

gboolean
clue_equal (ClueId a, ClueId b)
{
  return (a.direction == b.direction
          && a.index == b.index);
}

static gint
cell_array_cmp (gconstpointer a,
                gconstpointer b)
{
  IPuzCellCoord *coord_a = (IPuzCellCoord *) a;
  IPuzCellCoord *coord_b = (IPuzCellCoord *) b;

  if (coord_a->column == coord_b->column)
    return coord_a->row - coord_b->row;
  return coord_a->column - coord_b->column;
}

CellArray *
cell_array_new (void)
{
  return (CellArray *) g_array_new (FALSE, FALSE, sizeof (IPuzCellCoord));
}

CellArray *
cell_array_copy (CellArray *src)
{
  CellArray *dest;

  g_return_val_if_fail (src != NULL, NULL);

  dest = cell_array_new ();

  g_array_set_size (dest, cell_array_len (src));
  memcpy (((GArray*)dest)->data, ((GArray*)src)->data, cell_array_len (src) * sizeof (IPuzCellCoord));

  return dest;
}

/* Not a macro so we can pass it into other functions */
void
cell_array_unref (CellArray *cell_array)
{
  g_array_unref ((GArray *) cell_array);
}

void
cell_array_add (CellArray     *cell_array,
                IPuzCellCoord  coord)
{
  guint out;

  g_return_if_fail (cell_array != NULL);

  if (g_array_binary_search ((GArray *) cell_array,
                             &coord,
                             cell_array_cmp,
                             &out))
    return;

  g_array_append_val ((GArray *) cell_array, coord);
                             
}

void
cell_array_toggle (CellArray     *cell_array,
                   IPuzCellCoord  coord)
{
  guint out;

  g_return_if_fail (cell_array != NULL);

  if (g_array_binary_search ((GArray *) cell_array,
                             &coord,
                             cell_array_cmp,
                             &out))
    g_array_remove_index ((GArray *) cell_array, out);
  else
    {
      g_array_append_val ((GArray *) cell_array,
                          coord);
      g_array_sort ((GArray *) cell_array, cell_array_cmp);
    }
}

gboolean
cell_array_find (CellArray     *cell_array,
                 IPuzCellCoord  coord,
                 guint         *out)
{
  g_return_val_if_fail (cell_array != NULL, FALSE);

  return g_array_binary_search ((GArray *) cell_array,
                                &coord,
                                cell_array_cmp,
                                out);
}
