/* play-xword-column.c
 *
 * Copyright 2021 Jonathan Blandford <jrb@gnome.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-3.0-or-later
 */


#include "crosswords-config.h"
#include <glib/gi18n-lib.h>
#include <libipuz/libipuz.h>
#include "play-xword-column.h"


enum
{
  PROP_0,
  PROP_SPACING,
  N_PROPS
};

static GParamSpec *obj_props[N_PROPS] =  {NULL, };
static GtkBuildableIface *parent_buildable_iface;


struct _PlayXwordColumn
{
  GtkWidget parent_instance;

  guint spacing;
  GtkWidget *intro;
  GtkWidget *grid;
  GtkWidget *grid_swindow;
  GtkWidget *notes;
  GtkWidget *clue;
};


static void               play_xword_column_init                (PlayXwordColumn      *self);
static void               play_xword_column_class_init          (PlayXwordColumnClass *klass);
static void               play_xword_column_set_property        (GObject              *object,
                                                                 guint                 prop_id,
                                                                 const GValue         *value,
                                                                 GParamSpec           *pspec);
static void               play_xword_column_get_property        (GObject              *object,
                                                                 guint                 prop_id,
                                                                 GValue               *value,
                                                                 GParamSpec           *pspec);
static void               play_xword_column_dispose             (GObject              *object);
static void               play_xword_column_size_allocate       (GtkWidget            *widget,
                                                                 int                   width,
                                                                 int                   height,
                                                                 int                   baseline);
static void               play_xword_column_measure             (GtkWidget            *widget,
                                                                 GtkOrientation        orientation,
                                                                 int                   for_size,
                                                                 int                  *minimum,
                                                                 int                  *natural,
                                                                 int                  *minimum_baseline,
                                                                 int                  *natural_baseline);
static GtkSizeRequestMode play_xword_column_get_request_mode    (GtkWidget            *widget);
static void               play_xword_column_buildable_init      (GtkBuildableIface    *iface);
static void               play_xword_column_buildable_add_child (GtkBuildable         *buildable,
                                                                 GtkBuilder           *builder,
                                                                 GObject              *child,
                                                                 const char           *type);


G_DEFINE_TYPE_WITH_CODE (PlayXwordColumn, play_xword_column, GTK_TYPE_WIDGET,
                         G_IMPLEMENT_INTERFACE (GTK_TYPE_BUILDABLE,
                                                play_xword_column_buildable_init));


static void
play_xword_column_init (PlayXwordColumn *self)
{
  self->spacing = 16;
}

static void
play_xword_column_class_init (PlayXwordColumnClass *klass)
{
  GObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = G_OBJECT_CLASS (klass);
  widget_class = GTK_WIDGET_CLASS (klass);

  object_class->set_property = play_xword_column_set_property;
  object_class->get_property = play_xword_column_get_property;
  object_class->dispose = play_xword_column_dispose;
  widget_class->size_allocate = play_xword_column_size_allocate;
  widget_class->measure = play_xword_column_measure;
  widget_class->get_request_mode = play_xword_column_get_request_mode;

  obj_props[PROP_SPACING] = g_param_spec_uint ("spacing",
                                               "Spacing",
                                               "The space to put between the crossword elements",
                                               0, G_MAXUINT, 0,
                                               G_PARAM_READWRITE);
  g_object_class_install_properties (object_class, N_PROPS, obj_props);
}

static void
play_xword_column_set_property (GObject      *object,
                                guint         prop_id,
                                const GValue *value,
                                GParamSpec   *pspec)
{
  PlayXwordColumn *self = PLAY_XWORD_COLUMN (object);

  switch (prop_id)
    {
    case PROP_SPACING:
      self->spacing = g_value_get_uint (value);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
play_xword_column_get_property (GObject    *object,
                                guint       prop_id,
                                GValue     *value,
                                GParamSpec *pspec)
{
  PlayXwordColumn *self = PLAY_XWORD_COLUMN (object);

  switch (prop_id)
    {
    case PROP_SPACING:
      g_value_set_uint (value, self->spacing);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
    }
}

static void
play_xword_column_dispose (GObject *object)
{
  PlayXwordColumn *self = PLAY_XWORD_COLUMN (object);

  g_clear_pointer (&self->intro, gtk_widget_unparent);
  g_clear_pointer (&self->grid_swindow, gtk_widget_unparent);
  g_clear_pointer (&self->notes, gtk_widget_unparent);
  g_clear_pointer (&self->clue, gtk_widget_unparent);

  G_OBJECT_CLASS (play_xword_column_parent_class)->dispose (object);
}

static void
label_relative_allocation (GtkWidget *label,
                           gint       target_width,
                           gint      *height,
                           gint      *visible_children)
{
  if (gtk_widget_should_layout (label))
    {
      gint child_min, child_min_baseline, child_nat_baseline;
      gtk_widget_measure (label, GTK_ORIENTATION_VERTICAL, target_width,
                          &child_min, height,
                          &child_min_baseline, &child_nat_baseline);
      *visible_children += 1;
    }
}

static gboolean
allocate_widget (GtkWidget *widget,
                 gint      *y,
                 gint       width,
                 gint       height,
                 gint       spacing)
{
  GtkAllocation allocation = {
    .x = 0,
    .y = *y,
    .width = width,
    .height = height,
  };

  if (gtk_widget_should_layout (widget))
    {
      gtk_widget_size_allocate (widget, &allocation, -1);
      *y += height + spacing;
      return TRUE;
    }
  return FALSE;
}

static void
play_xword_column_size_allocate (GtkWidget *widget,
                                 int        width,
                                 int        height,
                                 int        baseline)
{
  PlayXwordColumn *self;
  gint grid_width;
  gint grid_height;
  gint child_nat = 0;
  gint child_min_baseline = -1;
  gint child_nat_baseline = -1;
  gint visible_children = 0;
  gint swindow_min = 0;
  gint intro_height = 0;
  gint notes_height = 0;
  gint swindow_height = 0;
  gint clue_height = 0;
  gint y = 0;

  self = PLAY_XWORD_COLUMN (widget);

  //g_print ("allocate: width %d   height %d\n", width, height);
  g_assert (self->intro != NULL && self->grid != NULL &&
            self->notes != NULL && self->clue != NULL);

  /* get the size of the grid. we calculate our overage based on that. */
  gtk_widget_measure (self->grid, GTK_ORIENTATION_HORIZONTAL, -1,
                      &grid_width, &child_nat,
                      &child_min_baseline, &child_nat_baseline);
  gtk_widget_measure (self->grid, GTK_ORIENTATION_VERTICAL, -1,
                      &grid_height, &child_nat,
                      &child_min_baseline, &child_nat_baseline);
  gtk_widget_measure (self->grid_swindow, GTK_ORIENTATION_VERTICAL, -1,
                      &swindow_min, &child_nat,
                      &child_min_baseline, &child_nat_baseline);
  visible_children ++; /* For the grid */

  /* Measure the labels */
  label_relative_allocation (self->intro, grid_width, &intro_height, &visible_children);
  label_relative_allocation (self->notes, grid_width, &notes_height, &visible_children);
  label_relative_allocation (self->clue, grid_width, &clue_height, &visible_children);
  /* Measure the rest of the widgets. Should always be visible, but we
   * test just in case*/

  /* allocate the scrolled window. */
  swindow_height = height - (intro_height + notes_height + clue_height + (visible_children-1)*self->spacing);
  swindow_height = MIN (swindow_height, grid_height);
  swindow_height = MAX (swindow_height, swindow_min);

  allocate_widget (self->intro, &y, width, intro_height, self->spacing);
  allocate_widget (self->grid_swindow, &y, width, swindow_height, self->spacing);
  allocate_widget (self->notes, &y, width, notes_height, self->spacing);
  allocate_widget (self->clue, &y, width, clue_height, self->spacing);
}

static void
play_xword_column_measure (GtkWidget      *widget,
                           GtkOrientation  orientation,
                           int             for_size,
                           int            *minimum,
                           int            *natural,
                           int            *minimum_baseline,
                           int            *natural_baseline)
{
  PlayXwordColumn *self;
  gint child_min = 0;
  gint child_nat = 0;
  gint child_min_baseline = -1;
  gint child_nat_baseline = -1;
  gint target_width;
  gint visible_children = 0;

  self = PLAY_XWORD_COLUMN (widget);

  g_assert (self->intro != NULL && self->grid != NULL &&
            self->notes != NULL && self->clue != NULL);

  *minimum = 0;
  *natural = 0;

  /* get the width of the grid. we calculate our height for wrapped
   * text based on that, instead of the height we're given. */
  gtk_widget_measure (self->grid, GTK_ORIENTATION_HORIZONTAL, -1,
                      &child_min, &child_nat,
                      &child_min_baseline, &child_nat_baseline);
  target_width = child_min;

  /* If we're measuring the horizontal, the grid size is all we
   * need */
  if (orientation == GTK_ORIENTATION_HORIZONTAL)
    {
      *minimum = target_width;
      *natural = target_width;
      return;
    }

  for (GtkWidget *child = gtk_widget_get_first_child (widget);
       child != NULL;
       child = gtk_widget_get_next_sibling (child))
    {
      if (gtk_widget_should_layout (child))
        {
          gtk_widget_measure (child, orientation, target_width,
                              &child_min, &child_nat,
                              &child_min_baseline, &child_nat_baseline);

          *minimum += child_min;
          *natural += child_nat;
          visible_children ++;

          if (child_min_baseline > -1)
            *minimum_baseline = MAX (*minimum_baseline, child_min_baseline);
          if (child_nat_baseline > -1)
            *natural_baseline = MAX (*natural_baseline, child_nat_baseline);
        }
    }

  *minimum += (visible_children - 1)*self->spacing;
  *natural += (visible_children - 1)*self->spacing;

  //g_print ("measure: minimum: %d natural: %d\n", *minimum, *natural);
}

static GtkSizeRequestMode
play_xword_column_get_request_mode (GtkWidget *widget)
{
  return GTK_SIZE_REQUEST_CONSTANT_SIZE;
}

static void
play_xword_column_buildable_init (GtkBuildableIface *iface)
{
  parent_buildable_iface = g_type_interface_peek_parent (iface);

  iface->add_child = play_xword_column_buildable_add_child;
}


static void
set_intro_widget (PlayXwordColumn *self,
                  GtkWidget       *intro)
{
  if (self->intro == intro)
    return;

  if (self->intro)
    gtk_widget_unparent (self->intro);

  self->intro = intro;

  if (intro)
    {
      gtk_widget_set_parent (self->intro, GTK_WIDGET (self));
    }
}

static void
set_grid_widget (PlayXwordColumn *self,
                 GtkWidget       *grid)
{
  if (self->grid == grid)
    return;

  if (self->grid_swindow)
    g_clear_pointer (&self->grid_swindow, gtk_widget_unparent);
  self->grid = grid;

  if (grid)
    {
      /* Put this in a scrolled window */
      self->grid_swindow =
        (GtkWidget *) g_object_new (GTK_TYPE_SCROLLED_WINDOW,
                                    "valign", GTK_ALIGN_FILL,
                                    "vexpand", TRUE,
                                    "vscrollbar-policy", GTK_POLICY_AUTOMATIC,
                                    "hscrollbar-policy", GTK_POLICY_NEVER,
                                    NULL);
      gtk_scrolled_window_set_child (GTK_SCROLLED_WINDOW (self->grid_swindow), self->grid);
      gtk_widget_set_parent (self->grid_swindow, GTK_WIDGET (self));
    }
}

static void
set_notes_widget (PlayXwordColumn *self,
                  GtkWidget       *notes)
{
  if (self->notes == notes)
    return;

  if (self->notes)
    gtk_widget_unparent (self->notes);
  self->notes = notes;

  if (notes)
    {
      gtk_widget_set_parent (self->notes, GTK_WIDGET (self));
    }
}

static void
set_clue_widget (PlayXwordColumn *self,
                 GtkWidget       *clue)
{
  if (self->clue == clue)
    return;

  if (self->clue)
    gtk_widget_unparent (self->clue);
  self->clue = clue;

  if (clue)
    {
      gtk_widget_set_parent (self->clue, GTK_WIDGET (self));
    }
}

static void
play_xword_column_buildable_add_child (GtkBuildable *buildable,
                                       GtkBuilder   *builder,
                                       GObject      *child,
                                       const char   *type)
{
  PlayXwordColumn *self;

  self = PLAY_XWORD_COLUMN (buildable);

  if (type && strcmp (type, "intro") == 0)
    set_intro_widget (self, GTK_WIDGET (child));
  else if (type && strcmp (type, "grid") == 0)
    set_grid_widget (self, GTK_WIDGET (child));
  else if (type && strcmp (type, "notes") == 0)
    set_notes_widget (self, GTK_WIDGET (child));
  else if (type && strcmp (type, "clue") == 0)
    set_clue_widget (self, GTK_WIDGET (child));
  else
    {
      parent_buildable_iface->add_child (buildable, builder, child, type);
    }
}



